
/* kit/grm/cmm.g */

/* --------------------------------------------------------------------- */

< expression CmmExpr >
< struct CmmExpr { enum kind; } >
< struct CmmBinaryExpr : CmmExpr {  } >
< struct CmmStat { enum stat; } >

/*
identifier
number
real_number
char_literal
string_literal
*/

/* --------------------------------------------------------------------- */

/* compound name */

base_name <modify CmmName>:
  (
     spec_func: special_function
  |
     id: identifier
  );

base_args <modify CmmName>:
  (
    (template_arg_list) =>
       template_args:template_arg_list
  )? ;

cont_item <CmmContName>:
  <no_space> "::" <no_space>
  <modify> cont_name
  <modify> cont_args ;

cont_name <modify CmmContName>:
  (
     '~' <no_space> <set a_destructor = true>
  |
     "template" <set a_template = true>
  )?
  (
     spec_func: special_function
  |
     id: identifier
  );

cont_args <modify CmmContName>:
  (
     (template_arg_list) =>
        template_args:template_arg_list
  )? ;

/* simple name */

simple_name <CmmName>:
  id: identifier; // only identifier

/* qualified name */

qualified_name <CmmName>:
  ( "::" <set a_global = true> )?
  <modify> base_name
  <modify> base_args
  ( <add> cont_item )*
  ;

/* --------------------------------------------------------------------- */

/* identifier expr */

ident_expr <CmmIdentExpr:CmmExpr, kind=identExp>:
   name:qualified_name
   <execute on_ident_expr> ;

/* primary expression */

primary_expr <select CmmExpr>:
    int_value_expr
  | real_value_expr
  | char_value_expr
  | string_value_expr
  | this_expr
  | subexpr_expr
  | ident_expr // !? last - important
  ;

int_value_expr <CmmIntValue:CmmExpr, kind=intValueExp>:
  value:number;

real_value_expr <CmmRealValue:CmmExpr, kind=realValueExp>:
  value:real_number;

char_value_expr <CmmCharValue:CmmExpr, kind=charValueExp>:
  value:character_literal;

string_value_cont <CmmStringCont>:
  value:string_literal;

string_value_expr <CmmStringValue:CmmExpr, kind=stringValueExp>:
  value:string_literal
  ( <add> string_value_cont )* ; // !?

this_expr <CmmThisExpr:CmmExpr, kind=thisExp>:
  "this";

subexpr_expr <CmmSubexprExpr:CmmExpr, kind=subexprExp>:
  '(' param:expr ')';

/* postfix expression */

postfix_start <select CmmExpr>:
    modern_cast_expr
  | typeid_expr
  | typename_expr // !?
  | type_change_expr // !?
  | primary_expr ; // !? last - important

modern_cast_expr <CmmModernCastExpr:CmmExpr>:
  ( "dynamic_cast"       <set kind = DynamicCast>
  | "static_cast"        <set kind = StaticCast>
  | "const_cast"         <set kind = ConstCast>
  | "reinterpret_cast"   <set kind = ReinterpreterCast> )
  '<' type:type_id '>'
  '(' param:expr ')'
  ;

typeid_expr <CmmTypeidExpr:CmmExpr, kind=typeidExp>:
  "typeid"
  '('
  (
     (type_id) =>
        type:type_id
  |
     value:expr
  )
  ')'
  ;

typename_expr <CmmTypenameExpr:CmmExpr, kind=typenameExp>:
  "typename"
  name:qualified_name
  '(' list:expr_list ')' ;

type_change_expr <CmmTypechangeExpr:CmmExpr, kind=typechangeExp>: // !?
  (
    "signed"   <set a_signed = true>   |
    "unsigned" <set a_unsigned = true> |

    "short"    <set a_short=true> |
    "long"     <set a_long=true>  ( "long" <set a_long_long=true> )? |

    "bool"     <set a_bool = true>     |
    "char"     <set a_char = true>     |
    "wchar_t"  <set a_wchar = true>    |
    "int"      <set a_int = true>      |
    "float"    <set a_float = true>    |
    "double"   <set a_double = true>   |
    "void"     <set a_void = true>
  )
  '(' list:expr_list ')' ;

postfix_expr <choose CmmExpr>:
  postfix_start
  (
    index_expr |
    call_expr |
    compound_expr |
    field_expr |
    ptr_field_expr |
    post_inc_expr |
    post_dec_expr
  )* ;

index_expr <CmmIndexExpr:CmmExpr, kind=indexExp>:
  <store left:CmmExpr>
  '[' param:expr ']' ;

call_expr <CmmCallExpr:CmmExpr, kind=callExp>:
  <store left:CmmExpr>
  '(' param_list:expr_list ')' ;

compound_expr <CmmCompoundExpr:CmmExpr, kind=compoundExp>:
  <store left:CmmExpr>
  '{' body:stat_list '}' ;

field_expr <CmmFieldExpr:CmmExpr, kind=fieldExp>:
  <store left:CmmExpr>
  <no_space> '.' <no_space>
  name:qualified_name
  <execute on_field_expr> ;

ptr_field_expr <CmmPtrFieldExpr:CmmExpr, kind=ptrFieldExp>:
  <store left:CmmExpr>
  <no_space> "->" <no_space>
  name:qualified_name
  <execute on_field_expr> ;

post_inc_expr <CmmPostIncExpr:CmmExpr, kind=postIncExp>:
  <store left:CmmExpr>
  "++" ;

post_dec_expr <CmmPostDecExpr:CmmExpr, kind=postDecExp>:
  <store left:CmmExpr>
  "--" ;

expr_list <CmmExprList>:
  (
     <add> assignment_expr
     ( ',' <add> assignment_expr )*
  )? ;

/* unary expression */

unary_expr <select CmmExpr>:
    inc_expr
  | dec_expr
  | deref_expr
  | addr_expr
  | plus_expr
  | minus_expr
  | bit_not_expr
  | log_not_expr
  | scope_expr
  | sizeof_expr
  | allocation_expr
  | deallocation_expr
  | throw_expr
  | postfix_expr ; // !? last - important

inc_expr <CmmUnaryExpr:CmmExpr, kind=incExp>:
  "++" param:cast_expr;

dec_expr <CmmUnaryExpr:CmmExpr, kind=decExp>:
  "--" param:cast_expr;

deref_expr <CmmUnaryExpr:CmmExpr, kind=derefExp>:
  '*' param:cast_expr;

addr_expr <CmmUnaryExpr:CmmExpr, kind=addrExp>:
  '&' param:cast_expr;

plus_expr <CmmUnaryExpr:CmmExpr, kind=plusExp>:
  '+' param:cast_expr;

minus_expr <CmmUnaryExpr:CmmExpr, kind=minusExp>:
  '-' param:cast_expr;

bit_not_expr <CmmUnaryExpr:CmmExpr, kind=bitNotExp>:
  '~' param:cast_expr;

log_not_expr <CmmUnaryExpr:CmmExpr, kind=logNotExp>:
  '!' param:cast_expr;

scope_expr <CmmUnaryExpr:CmmExpr, kind=scopeExp>: // !?
   "::" param:cast_expr ;

sizeof_expr <CmmSizeofExp:CmmExpr, kind=sizeofExp>:
  "sizeof"
  (
     ( '(' type_id ')' ) =>
       '(' type:type_id ')'
  |
     value:unary_expr
  );

/* new */

allocation_expr <CmmNewExpr:CmmExpr, kind=newExp>:
  // ( "::" <set a_global = true> )?
  "new"
  // (
  //    placement:new_placement
  // )?
  (
     type1:new_type_id
  |
    '(' type2:type_id ')'
  )
  (
    '(' init:expr_list ')'
  )? ;

// new_placement <CmmNewPlacement>:
//   '(' list:expr_list ')' ;

new_type_id <CmmNewTypeId>:
  type_spec:type_specifiers
  ptr:ptr_specifier_list
  ( <add> allocation_array_limit )* ;

allocation_array_limit <CmmNewArrayLimit>:
  '[' value:expr ']'  ;

/* delete */

deallocation_expr <CmmDeleteExpr:CmmExpr, kind=deleteExp>:
  // ( "::" <set a_global = true> )?
  "delete"
  ( '[' ']' <set a_array = true> )?
  param:cast_expr ;

/* cast expression */

cast_expr <select CmmExpr>:
  //   ( cast_formula ) => // !?
  //      cast_formula
  // |
    unary_expr ;

// cast_formula <CmmCastFormula: CmmExpr>:
//   '(' type:type_id ')' param:cast_expr ;

/* pm expression */

pm_expr <choose CmmExpr>:
  cast_expr
  (
    <new CmmPtrMemExpr:CmmExpr>
    <store left:CmmExpr>
    ( ".*"  <set kind=dotMemberExp> |
      "->*" <set kind=arrowMemberExp> )
    right:cast_expr
  )* ;

/* arithmetic and logical expressions */

multiplicative_expr <choose CmmExpr>:
  pm_expr
  (
    <new CmmMulExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    ( '*' <set kind=mulExp> |
      '/' <set kind=divExp> |
      '%' <set kind=modExp> )
    right:pm_expr
    <execute on_binary_expr>
  )* ;

additive_expr <choose CmmExpr>:
  multiplicative_expr
  (
    <new CmmAddExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    ( '+' <set kind=addExp> |
      '-' <set kind=subExp> )
    right:multiplicative_expr
    <execute on_binary_expr>
  )* ;

shift_expr <choose CmmExpr>:
  additive_expr
  (
    <new CmmShiftExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    ( "<<" <set kind=shlExp> |
      ">>" <set kind=shrExp> )
    right:additive_expr
  )* ;

relational_expr <choose CmmExpr>:
  shift_expr
  (
    <new CmmRelExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    ( '<'  <set kind=ltExp> |
      '>'  <set kind=gtExp> |
      "<=" <set kind=leExp> |
      ">=" <set kind=geExp> )
    right:shift_expr
  )* ;

equality_expr <choose CmmExpr>:
  relational_expr
  (
    <new CmmEqExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    ( "==" <set kind=eqExp> |
      "!=" <set kind=neExp> )
    right:relational_expr
  )* ;

and_expr <choose CmmExpr>:
  equality_expr
  (
    <new CmmAndExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    '&' <set kind=bitAndExp>
    right:equality_expr
  )* ;

exclusive_or_expr <choose CmmExpr>:
  and_expr
  (
    <new CmmXorExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    '^' <set kind=bitXorExp>
    right:and_expr
  )* ;

inclusive_or_expr <choose CmmExpr>:
  exclusive_or_expr
  (
    <new CmmOrExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    ( '|' <set kind=bitOrExp> )
    right:exclusive_or_expr
  )* ;

logical_and_expr <choose CmmExpr>:
  inclusive_or_expr
  (
    <new CmmAndAndExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    "&&" <set kind=logAndExp>
    right:inclusive_or_expr
  )* ;

logical_or_expr <choose CmmExpr>:
  logical_and_expr
  (
    <new CmmOrOrExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    "||" <set kind=logOrExp>
    right:logical_and_expr
  )* ;

/* conditional and assignment expression */

assignment_expr <choose CmmExpr>:
  logical_or_expr // !?
  (
    <new CmmAssignExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    ( '='   <set kind=assignExp>    |
      "+="  <set kind=addAssignExp> |
      "-="  <set kind=subAssignExp> |
      "*="  <set kind=mulAssignExp> |
      "/="  <set kind=divAssignExp> |
      "%="  <set kind=modAssignExp> |
      "<<=" <set kind=shlAssignExp> |
      ">>=" <set kind=shrAssignExp> |
      "&="  <set kind=andAssignExp> |
      "^="  <set kind=xorAssignExp> |
      "|="  <set kind=orAssignExp>  |
      '?'   <set kind=condExp> middle:expr ':' )
    right:assignment_expr
  )? ;

conditional_expr <return CmmExpr>:
  assignment_expr ;

/* expression */

expr <choose CmmExpr>:
  assignment_expr
  (
    <new CmmCommaExpr:CmmBinaryExpr>
    <store left:CmmExpr>
    ( ',' <set kind=commaExp> )
    right:assignment_expr
  )* ;

/* constant expression */

const_expr <return CmmExpr>:
  conditional_expr ;

/* --------------------------------------------------------------------- */

/* statement */

stat <select CmmStat>:

     ( identifier ':' ) => labeled_stat |
     ( block_declaration_head ) => declaration_stat |
     expression_stat |
     empty_stat |

     compound_stat |
     case_stat |
     default_stat |
     if_stat |
     switch_stat |
     while_stat |
     do_stat |
     for_stat |
     break_stat |
     continue_stat |
     return_stat |
     goto_stat |
     try_stat;

/* inner_stat - statements with different output indentation */

inner_stat <return CmmStat>:
   <new_line>
   <indent>
   stat
   <unindent> ;

/* list of statements */

stat_list <CmmStatSect>:
   ( <new_line> <add> stat )* ;

/* conditions */

condition <CmmCondition>:
     ( condition_declaration ')' ) =>
        cond_decl:condition_declaration
  |
     cond_expr:expr
  ;

for_condition <CmmCondition>:
     ( condition_declaration ';' ) =>
        cond_decl:condition_declaration
  |
     cond_expr:expr
  ;

/* concrete statements */

declaration_stat <CmmDeclStat:CmmStat, stat=declarationStat>:
  inner_decl:block_declaration ;

labeled_stat <CmmLabeledStat:CmmStat, stat=labeledStat>:
  lab:identifier ':'
  <new_line>
  body:stat ;

case_stat <CmmCaseStat:CmmStat, stat=caseStat>:
  "case" case_expr:expr ':'
  <new_line>
  body:stat;

default_stat <CmmDefaultStat:CmmStat, stat=defaultStat>:
  "default" ':'
  <new_line>
  body:stat;

expression_stat <CmmExprStat:CmmStat, stat=expressionStat>:
  inner_expr:expr ';';

empty_stat <CmmEmptyStat:CmmStat, stat=emptyStat>:
  ';' ;

compound_stat <CmmCompoundStat:CmmStat, stat=compoundStat>:
  '{'
  <indent>
  body:stat_list
  <unindent>
  <new_line>
  '}' ;

if_stat <CmmIfStat:CmmStat, stat=ifStat>:
  "if" '(' cond:condition ')' then_stat:inner_stat
  (
    <silent>
    <new_line>
    "else"
     else_stat:inner_stat
  )? ;

switch_stat <CmmSwitchStat:CmmStat, stat=switchStat>:
  "switch" '(' cond:condition ')' body:inner_stat ;

while_stat <CmmWhileStat:CmmStat, stat=whileStat>:
  "while" '(' cond:condition ')' body:inner_stat ;

do_stat <CmmDoStat:CmmStat, stat=doStat>:
  "do" body:inner_stat "while" '(' cond_expr:expr ')' ';' ;

for_stat <CmmForStat:CmmStat, stat=forStat>:
  "for"
  '('
  ( from_expr:for_condition )?
  ';'
  (to_expr:expr)?
  ';'
  (cond_expr:expr)?
  ')'
  body:inner_stat ;

break_stat <CmmBreakStat:CmmStat, stat=breakStat>:
  "break" ';' ;

continue_stat <CmmContinueStat:CmmStat, stat=continueStat>:
  "continue" ';' ;

return_stat <CmmReturnStat:CmmStat, stat=returnStat>:
  "return" (return_expr:expr)? ';' ;

goto_stat <CmmGotoStat:CmmStat, stat=gotoStat>:
  "goto" goto_lab:identifier ';' ;

/* ---------------------------------------------------------------------- */

/* declaration */

declaration_list <CmmDeclSect>:
  ( <add> declaration <new_line> )* ;

declaration <select CmmDecl>:

    ( constructor_head ) => constructor_declaration
  | simple_declaration
  | empty_declaration

  | class_declaration
  | enum_declaration

  | typedef_declaration
  | friend_declaration

  | namespace_declaration
  | extern_declaration
  | using_declaration

  | template_declaration
  ;

/* block declaration - used only in declaration_stat */

block_declaration <select CmmDecl>:

    simple_declaration
  | typedef_declaration
  | using_declaration
  ;

/* block declaration head - only for predicate */

block_declaration_head:
    simple_head
  | "typedef"
  | "using"
  ;

/* ---------------------------------------------------------------------- */

/* namespace declaration */

namespace_declaration <CmmNamespaceDecl:CmmDecl> :
  <execute open_namespace>
  "namespace"
  name:simple_name
  <execute on_namespace>
  <new_line>
  '{'
     <execute start_namespace>
     <new_line>
     <indent>
     body:declaration_list
     <unindent>
     <execute stop_namespace>
  '}'
  <execute close_namespace> ;

/* using declaration */

using_declaration <CmmUsingDecl:CmmDecl>:
  "using"
  (
     "namespace" <set a_namespace = true>
  |
     "typename"  <set a_typename  = true>
  )?
  name:qualified_name
  ';' ;

/* linkage declaration */

extern_declaration <CmmExternDecl:CmmDecl>:
  "extern"
  ( language:string_literal )?
  (
     '{' decl_list:declaration_list '}'
  |
     inner_decl:declaration
  ) ;

/* typedef declaration */

typedef_declaration <CmmTypedefDecl:CmmDecl>:
  "typedef"
  type_spec:type_specifiers
  inner_decl:declarator
  ';' ;

/* friend declaration */

friend_declaration <CmmFriendDecl:CmmDecl>:
  "friend"
  inner_decl:declaration ;

/* empty declaration */

empty_declaration <CmmEmptyDecl:CmmDecl>:
  ';' ;

/* ---------------------------------------------------------------------- */

/* declaration specifiers */

decl_specifiers <CmmDeclSpec>:
  (
    "inline"   <set a_inline = true>   |
    "virtual"  <set a_virtual = true>  |
    "explicit" <set a_explicit = true> |
    "mutable"  <set a_mutable = true>  |

    // "extern"   <set a_extern = true>   |
    "static"   <set a_static = true>   |
    // "auto"     <set a_auto = true>     |
    "register" <set a_register = true>
  )* ;

/* type specifiers */

type_specifiers <CmmTypeSpec>:
  (
    "const" <set a_const = true>
  |
    "volatile" <set a_volatile = true>
  )?
  (
    "signed" <set a_signed = true>
  |
    "unsigned" <set a_unsigned = true>
  )?
  (
     basic_name:qualified_name |

     "typename" typename_name:qualified_name |

     "short"    <set a_short=true> |
     "long"     <set a_long=true>  ( "long" <set a_long_long=true> )? |

     "bool"     <set a_bool = true>     |
     "char"     <set a_char = true>     |
     "wchar_t"  <set a_wchar = true>    |
     "int"      <set a_int = true>      |
     "float"    <set a_float = true>    |
     "double"   <set a_double = true>   |
     "void"     <set a_void = true>
  )
  <execute on_type_specifiers> ;

/* ---------------------------------------------------------------------- */

/* pointer specifiers */

ptr_specifier_list <CmmPtrSpecifierSect>:
  ( <add> ptr_specifier )* ;

ptr_specifier <CmmPtrSpecifier>:
    '*' <set pointer=true>
    <modify> ptr_cv_specifier_list
  |
    '&' <set reference=true>
    <modify> ptr_cv_specifier_list
  // |
  //   (type_name "::" '*') =>
  //   name:type_name "::" '*'
  //   <modify> ptr_cv_specifier_list // !?
  ;

ptr_cv_specifier_list <modify CmmPtrSpecifier>:
  (
    "const" <set cv_const = true>
  |
    "volatile" <set cv_volatile = true>
  )* ;

/* function and array specifiers */

cont_specifier_list <CmmContSpecifierSect>:
  ( <add> cont_specifier )* ;

cont_specifier <select CmmContSpecifier>:
    function_specifier
  |
    array_specifier ;

function_specifier <CmmFunctionSpecifier:CmmContSpecifier>:
  '('
  parameters:parameter_declaration_list
  ')'
  (
    "const" <set cv_const = true>
  |
    "volatile" <set cv_volatile = true>
  )*
  ( exception_spec:exception_specification )? ;

array_specifier <CmmArraySpecifier:CmmContSpecifier>:
  '['
  (lim:expr)?
  ']' ;

/* parameter declarations */

parameter_declaration_list <CmmParamSect>:
  (
     <add> parameter_declaration
     ( ',' <add> parameter_declaration )*
  )? ;

parameter_declaration <CmmParamItem>:
    type:common_type_id
    ( '=' init:assignment_expr )?
  |
     "..." <set dots=true> // !?
  ;

/* ---------------------------------------------------------------------- */

/* simple declaration */

simple_declaration <CmmSimpleDecl:CmmDecl>:
  decl_spec:decl_specifiers
  type_spec:type_specifiers
  <add> simple_item
  <execute on_simple_item>
  (
     <execute open_function>
     func_body:function_body
     <execute close_function>
  |
     (
        ','
        <add> simple_item
        <execute on_simple_item>
     )*
     ';'
  );

simple_item <CmmSimpleItem>:
  decl:declarator
  ( ':' width:const_expr )?
  ( init:initializer )?
  ;

/* simple declaration head - only for predicate */

simple_head:
  decl_specifiers
  type_specifiers
  declarator
  ( '=' | '{' | ',' | ';' )
  ;

/* condition declaration */

condition_declaration <CmmSimpleDecl>:
  decl_spec:decl_specifiers
  type_spec:type_specifiers
  <add> simple_item
  ( ',' <add> simple_item )*
  // no semicolon
  ;

/* constructor declaration */

constructor_declaration <CmmCtorDecl:CmmSimpleDecl>:
  decl_spec:decl_specifiers
  <add> constructor_item
  (
     ( ':' <new_line> ctor_init:ctor_initializer )?
     func_body:function_body
  |
     '=' zero:number
  |
     ';'
  );

constructor_item <CmmSimpleItem>:
  ( '~' <set a_destructor = true> )?
  decl:declarator
  ;

/* constructor declaration head - only for predicate */

constructor_head:
  decl_specifiers
  ( '~' )?
  qualified_name
  cont_specifier_list
  ( ':' | '{' | '=' | ';' )
  ;

/* function body */

function_body <return CmmCompoundStat>:
  <new_line>
  compound_stat ;

/* ---------------------------------------------------------------------- */

// declarator          ... name or inner declarator
// abstract_declarator ... empty or inner declarator
// common_declarator   ... empty, name or inner declarator

/* ---------------------------------------------------------------------- */

/* declarator */

declarator <CmmDeclarator>:
  ptr:ptr_specifier_list
  (
     name:qualified_name
     <execute on_declarator>
  |
     ( nested_declarator ) => // necessary
        '(' inner:declarator ')'
  )
  cont:cont_specifier_list
  ;

abstract_declarator <CmmDeclarator>:
  ptr:ptr_specifier_list
  (
     ( nested_declarator ) =>
        '(' inner:abstract_declarator ')'
  )?
  cont:cont_specifier_list
  ;

common_declarator <CmmDeclarator>:
  ptr:ptr_specifier_list
  (
     name:qualified_name
     <execute on_declarator>
  |
     ( nested_declarator ) =>
        '(' inner:common_declarator ')'
  )?
  cont:cont_specifier_list
  ;

nested_declarator <CmmNestedDeclarator>:
  ( '(' <set something=true> )@
  ( '*' <set a_ptr = true> |
    '&' <set a_ref = true> );
  // member_pointer:simple_name "::" '*'

/* ---------------------------------------------------------------------- */

// type_id ... used in expressions, initialization of template_type_param, template_type_arg, exception_specification
// common_type_id ... used in parameter_declaration, template_value_param, try handler

/* ---------------------------------------------------------------------- */

/* type id */

type_id <CmmTypeId>: /* abstract_type_id */
  type_spec:type_specifiers
  decl:abstract_declarator ;

common_type_id <CmmTypeId>:
  type_spec:type_specifiers
  decl:common_declarator ;

/* ---------------------------------------------------------------------- */

/* initializer */

initializer <CmmInitializer>:
   '=' value:initializer_item ;

initializer_item <select CmmInitItem>:
     simple_initializer
   | initializer_list ;

simple_initializer <CmmInitSimple:CmmInitItem>:
  expr:assignment_expr ;

initializer_list <CmmInitList:CmmInitItem>:
  '{'
  <indent>
  (
    <add> initializer_item
    (
      ','
      <new_line>
      <add> initializer_item
    )*
  )?
  <unindent>
  '}';

/* ---------------------------------------------------------------------- */

/* enum */

enum_declaration <CmmEnumDecl:CmmDecl>:
  <execute open_enum>
  "enum"
  name:qualified_name
  <execute on_enum>
  (
     <new_line>
     '{'
        <execute start_enum>
        <new_line>
        <indent>
        enum_items:enum_list
        <unindent>
        <new_line>
        <execute stop_enum>
     '}'
   )?
   ';'
   <execute close_enum> ;

enum_list <CmmEnumSect>:
  (
    <add> enumerator
    ( ',' <add> enumerator )*
  )? ;

enumerator <CmmEnumItem>:
  <new_line>
  name:simple_name
  <execute on_enum_item>
  ( '=' init:const_expr )? ;

/* ---------------------------------------------------------------------- */

/* class */

class_declaration <CmmClassDecl:CmmDecl>:
  <execute open_class>
  (
    "class"  <set style = ClassStyle> |
    "struct" <set style = StructStyle> |
    "union"  <set style = UnionStyle>
  )
  name:qualified_name
  <execute on_class>
  (
     ( ':' base_list:base_specifier_list )?
     <new_line>
    '{'
        <execute start_class>
        <new_line>
        <indent>
        members:member_list
        <unindent>
        <execute stop_class>
    '}'
  )?
  ';'
  <execute close_class> ;

/* members */

member_visibility <CmmMemberVisibility:CmmDecl>:
  ( "private"   <set access = PrivateAccess>
  | "protected" <set access = ProtectedAccess>
  | "public"    <set access = PublicAccess> )
  <no_space> ':' ;

member_item <select CmmDecl>:

    member_visibility

  | ( constructor_head ) => constructor_declaration
  | simple_declaration
  | empty_declaration

  | class_declaration
  | enum_declaration

  | typedef_declaration
  | friend_declaration

  | using_declaration
  | template_declaration ;

member_list <CmmDeclSect>:
  ( <add> member_item <new_line> )* ;

/* base specification */

base_specifier_list <CmmBaseSect>:
  <add> base_specifier
  ( ',' <add> base_specifier )* ;

base_specifier <CmmBaseItem>:
  ( "virtual" <set a_virtual = true> )?
  ( "private"   <set access = PrivateAccess>
  | "protected" <set access = ProtectedAccess>
  | "public"    <set access = PublicAccess>
  |             <set access = NoAccess> )
  from_cls:qualified_name ;

/* constructor initializer */

ctor_initializer <CmmCtorInitializer>:
  <indent>
  <add> member_initializer
  (
     ','
     <new_line>
     <add> member_initializer
  )*
  <unindent> ;

member_initializer <CmmMemberInitializer>:
   name:qualified_name // !? qualified_name
   '('
   params:expr_list
   ')';

/* ---------------------------------------------------------------------- */

/* special member functions */

conversion_specifiers <return CmmTypeSpec>: // !?
  type_specifiers
  ptr:ptr_specifier_list ;

special_function <CmmSpecialFuction>:
   "operator"
   (
     conv:conversion_specifiers // !?

   | "new"    <set spec_new=true>    ( ('[' ']') => '[' ']' <set spec_new_array=true> )? // !?
   | "delete" <set spec_delete=true> ( ('[' ']') => '[' ']' <set spec_delete_array=true> )?

   | "+" <set spec_add=true>
   | "-" <set spec_sub=true>
   | "*" <set spec_mul=true>
   | "/" <set spec_div=true>
   | "%" <set spec_mod=true>
   | "^" <set spec_xor=true>
   | "&" <set spec_and=true>
   | "|" <set spec_or=true>
   | "~" <set spec_not=true>

   | "!"  <set spec_log_not=true>
   | "="  <set spec_assign=true>
   | "<"  <set spec_lt=true>
   | ">"  <set spec_gt=true>
   | "+=" <set spec_add_assign=true>
   | "-=" <set spec_sub_assign=true>
   | "*=" <set spec_mul_assign=true>
   | "/=" <set spec_div_assign=true>
   | "%=" <set spec_mod_assign=true>

   | "^="  <set spec_xor_assign=true>
   | "&="  <set spec_and_assign=true>
   | "|="  <set spec_or_assign=true>
   | "<<"  <set spec_shl=true>
   | ">>"  <set spec_shr=true>
   | ">>=" <set spec_shl_assign=true>
   | "<<=" <set spec_shr_assign=true>
   | "=="  <set spec_eq=true>
   | "!="  <set spec_ne=true>

   | "<="  <set spec_le=true>
   | ">="  <set spec_ge=true>
   | "&&"  <set spec_log_and=true>
   | "||"  <set spec_log_or=true>
   | "++"  <set spec_inc=true>
   | "--"  <set spec_dec=true>
   | ","   <set spec_comma=true>
   | "->*" <set spec_member_deref=true>
   | "->"  <set spec_deref=true>

   | "(" ")" <set spec_call=true>
   | "[" "]" <set spec_index=true>
   ) ;

/* ---------------------------------------------------------------------- */

/* template declaration */

template_declaration <CmmTemplateDecl:CmmDecl>:
  (
    "export"
    <set export_template = true>
  )?
  "template"
  (
    '<'
    params:template_param_list
    '>'
  )?
  inner_decl:declaration
  <execute on_template_declaration>
  ;

template_param_list <CmmTemplateParamSect>:
  (
    <add> template_param
    ( ',' <add> template_param )*
  )?;

template_param <select CmmTemplateParam>:
    ( "template" | "class" | "typename" ) => template_type_param
  |
    template_value_param
  ;

template_type_param <CmmTemplateTypeParam:CmmTemplateParam>:
  (
    "template" '<' params:template_param_list '>'
  )?
  (
    "class" <set a_class = true>
  |
    "typename" <set a_typename = true>
  )
  ( name:simple_name )?
  ( '=' init_type:type_id )?
  ;

template_value_param <CmmTemplateValueParam:CmmTemplateParam>:
  type:common_type_id
  ( '=' init_value:shift_expr )? ; // !?

/* template arguments */

template_arg_list <CmmTemplateArgSect>:
  '<'
  (
    <add> template_arg
    ( ',' <add> template_arg )*
  )?
  '>' ;

template_arg <select CmmTemplateArg>:
    (type_id) =>
      template_type_arg
  |
    template_value_arg;

template_type_arg <CmmTemplateTypeArg:CmmTemplateArg>:
  type:type_id; // !?

template_value_arg <CmmTemplateValueArg:CmmTemplateArg>:
   value:shift_expr; // !?

/* ---------------------------------------------------------------------- */

/* try statement */

try_stat <CmmTryStat:CmmStat, stat=tryStat>:
  "try"
  <new_line>
  body:compound_stat
  handlers:handler_list ;

handler_list <CmmHandlerSect>:
  <add> handler
  ( <add> handler )* ;

handler <CmmHandlerItem>:
  <new_line>
  "catch"
  '('
  (
    type:common_type_id
  |
    "..." <set dots=true>
  )
  ')'
  <new_line>
  body:compound_stat ;

/* throw expression */

throw_expr <CmmThrowExpr:CmmExpr, kind=throwExp >:
  "throw" (expr:assignment_expr)? ;

/* exception specification */

exception_specification <CmmExceptionSect>:
  "throw"
  '('
  (
    <add> exception_specification_item
    (',' <add> exception_specification_item)*
  )?
  ')' ;

exception_specification_item <CmmExceptionItem: CmmTypeDecl>:
  type:type_id ;

/* ---------------------------------------------------------------------- */
