#!/usr/bin/env python

from __future__ import print_function

import sys, os, re, traceback

use_qt5 = False
if use_qt5 :
   from PyQt5.QtCore import *
   from PyQt5.QtGui import *
   from PyQt5.QtWidgets import *
else :
   from PyQt4.QtCore import *
   from PyQt4.QtGui import *

sys.path.insert (1, os.path.join (os.getcwd (), "../grm"))
sys.path.insert (1, os.path.join (os.getcwd (), ".."))

from gcc_options import gcc_options, pkg_options

from lexer import Lexer, LexerException, fileNameToIndex

from util import Text

if sys.version_info >= (3,):
   def unicode (s) :
       return str (s)

# --------------------------------------------------------------------------

class FindBox (QWidget) :

   def __init__ (self, parent = None) :
       super (FindBox, self).__init__ (parent)

       self.edit = None
       self.small = False

       self.horiz_layout = QHBoxLayout ()
       self.setLayout (self.horiz_layout)

       self.closeButton = QToolButton ()
       self.closeButton.setIcon (QIcon.fromTheme ("dialog-close")) # !?
       self.closeButton.setShortcut ("Esc")
       # self.closeButton.setMaximumWidth (20)
       # self.closeButton.setFlat (True)
       self.closeButton.clicked.connect (self.hide)
       self.horiz_layout.addWidget (self.closeButton)

       self.label = QLabel ()
       self.label.setText ("Find:")
       self.horiz_layout.addWidget (self.label)

       # self.line = QLineEdit ()
       self.line = QComboBox ()
       self.line.setEditable (True)
       # self.line.setFrame (False)
       # self.line.setDuplicatesEnabled (False)
       # self.line.setInsertPolicy (QComboBox.InsertAtBottom)
       self.horiz_layout.addWidget (self.line)

       self.wholeWords = QCheckBox ()
       self.wholeWords.setText ("Whole &words")
       self.horiz_layout.addWidget (self.wholeWords)

       self.matchCase = QCheckBox ()
       self.matchCase.setText ("Match &case")
       self.horiz_layout.addWidget (self.matchCase)

       self.nextButton = QPushButton ()
       self.nextButton.setText ("&Next")
       self.nextButton.clicked.connect (self.findNext)
       self.horiz_layout.addWidget (self.nextButton)

       self.prevButton = QPushButton ()
       self.prevButton.setText ("&Prev")
       self.prevButton.clicked.connect (self.findPrev)
       self.horiz_layout.addWidget (self.prevButton)

       self.horiz_layout.setStretch (2, 10)

       # self.line.returnPressed.connect (self.returnPressed)
       self.line.lineEdit().returnPressed.connect (self.returnPressed)

   def resizeEvent (self, e) :
       w = e.size().width()
       if w < 640 :
          if not self.small :
             self.wholeWords.setText ("&Words")
             self.matchCase.setText ("&Case")
             self.small = True
       else :
          if self.small :
             self.wholeWords.setText ("Whole &words")
             self.matchCase.setText ("Match &case")
             self.small = False
       super (FindBox, self).resizeEvent (e)

   def returnPressed (self) :
       self.findNext ()

   def open (self) :
       self.show ()
       self.line.setFocus ()

   def findNext (self) :
       self.find (False)

   def findPrev (self) :
       self.find (True)

   def find (self, back) :
       # text = self.line.text ()
       text = self.line.currentText ()
       if text != "" :
          if self.line.findText (text) == -1 :
             self.line.addItem (text)

          opt = QTextDocument.FindFlags (0)
          if self.wholeWords.isChecked () :
             opt = opt | QTextDocument.FindWholeWords
          if self.matchCase.isChecked () :
             opt = opt | QTextDocument.FindCaseSensitively
          if back :
             opt = opt | QTextDocument.FindBackward

          ok = self.edit.find (text, opt)
          if not ok :
             if back :
                self.edit.moveCursor (QTextCursor.End)
             else :
                self.edit.moveCursor (QTextCursor.Start)
             self.edit.find (text, opt)

# --------------------------------------------------------------------------

class GoToLineBox (QWidget) :

   def __init__ (self, parent=None) :
       super (GoToLineBox, self).__init__ (parent)

       self.edit = None

       self.horiz_layout = QHBoxLayout ()
       self.setLayout (self.horiz_layout)

       self.closeButton = QToolButton ()
       self.closeButton.setIcon (QIcon.fromTheme ("dialog-close")) # !?
       self.closeButton.setShortcut ("Esc")
       self.closeButton.clicked.connect (self.hide)
       self.horiz_layout.addWidget (self.closeButton)

       self.label = QLabel ()
       self.label.setText ("Go to line:")
       self.horiz_layout.addWidget (self.label)

       self.line = QSpinBox ()
       self.horiz_layout.addWidget (self.line)

       self.button = QPushButton ()
       self.button.setText ("Go")
       self.horiz_layout.addWidget (self.button)

       self.horiz_layout.addStretch ()

       self.line.editingFinished.connect (self.editingFinished)

   def open (self) :
       cursor = self.edit.textCursor ()
       document = self.edit.document ()
       self.line.setMinimum (1)
       self.line.setMaximum (document.blockCount ())
       self.line.setValue (cursor.blockNumber () + 1) # line number
       self.line.setFocus ()
       self.show ()

   def editingFinished (self) :
       self.edit.selectLine (self.line.value ())

# --------------------------------------------------------------------------

class TextEdit (QTextBrowser) :
   def __init__ (self, parent = None) :
       super (TextEdit, self).__init__ (parent)
       self.win = None

       if isinstance (self, QPlainTextEdit) :
          self.setLineWrapMode (QPlainTextEdit.NoWrap) # otherwise selectLine is not exact
       else :
          self.setLineWrapMode (QPlainTextEdit.NoWrap)

       if isinstance (self, QTextBrowser) :
          self.setOpenLinks (False)
          self.anchorClicked.connect (self.onAnchorClick)

       self.setMouseTracking (True)

       self.findBox = FindBox ()
       self.findBox.edit = self

       self.lineBox = GoToLineBox ()
       self.lineBox.edit = self

       self.enableJump = False

   def onAnchorClick (self, link) :
       if self.win != None :
          self.win.click (int (link.toString()))

   def findText (self) :
       self.lineBox.hide ()
       self.findBox.open ()

   def findNext (self) :
       self.findBox.findNext ()

   def findPrev (self) :
       self.findBox.findPrev ()

   def goToLine (self) :
       self.findBox.hide ()
       self.lineBox.open ()

   def selectLine (self, line) :
       cursor = self.textCursor ()
       cursor.movePosition (QTextCursor.Start)
       cursor.movePosition (QTextCursor.NextBlock, QTextCursor.MoveAnchor, line-1)
       cursor.movePosition (QTextCursor.EndOfLine, QTextCursor.KeepAnchor)
       self.setTextCursor (cursor)
       self.ensureCursorVisible ()

   def select (self, start_line, start_col, stop_line, stop_col) :
       cursor = self.textCursor ()
       cursor.movePosition (QTextCursor.Start)
       cursor.movePosition (QTextCursor.NextBlock, QTextCursor.MoveAnchor, start_line-1)
       cursor.movePosition (QTextCursor.NextCharacter, QTextCursor.MoveAnchor, start_col-1)

       if start_line == stop_line :
          cursor.movePosition (QTextCursor.NextCharacter, QTextCursor.KeepAnchor, stop_col-start_col)
       else :
          cursor.movePosition (QTextCursor.NextBlock, QTextCursor.KeepAnchor, stop_line-start_line)
          cursor.movePosition (QTextCursor.StartOfLine, QTextCursor.KeepAnchor)
          cursor.movePosition (QTextCursor.NextCharacter, QTextCursor.KeepAnchor, stop_col)

       self.setTextCursor (cursor)
       self.ensureCursorVisible ()

   def event (self, e) :
       if e.type() == QEvent.ToolTip :
          cursor = self.cursorForPosition (e.pos ())
          cursor.select (QTextCursor.WordUnderCursor)
          if cursor.selectedText() != "" :
             tooltip = None
             fmt = cursor.charFormat ()
             if fmt.hasProperty (Text.infoProperty) :
                tooltip = "info: " + fmt.stringProperty (Text.infoProperty)
             elif fmt.hasProperty (QTextFormat.TextToolTip) :
                tooltip = "tooltip: " + fmt.toolTip()
             if tooltip != None :
                QToolTip.showText (e.globalPos(), tooltip)
          else :
             QToolTip.hideText ()
          return True
       else :
          return super (TextEdit, self).event (e)

   def mouseMoveEvent (self, e) :
       mask = Qt.ShiftModifier | Qt.ControlModifier | Qt.AltModifier | Qt.MetaModifier
       mod = int (e.modifiers () & mask)
       if mod == Qt.ControlModifier :
          if not self.enableJump :
             cursor = self.cursorForPosition (e.pos ())
             cursor.select (QTextCursor.WordUnderCursor)
             if cursor.selectedText() != "" :
                fmt = cursor.charFormat ()
                if not fmt.fontUnderline () : # keep red underline
                   fmt.setFontUnderline (True)
                   fmt.setUnderlineColor (Qt.blue)
                   cursor.setCharFormat (fmt)
       super (TextEdit, self).mouseMoveEvent (e)

   def mousePressEvent (self, e) :
       cursor = self.cursorForPosition (e.pos ())
       if self.enableJump :
          cursor.select (QTextCursor.BlockUnderCursor)
          text = cursor.selectedText()
          text = unicode (text)
          text = text.strip ()
          text = str (text)
          # print ("JUMP", "(" + text + ")")
          self.jump (text)
       else :
          mask = Qt.ShiftModifier | Qt.ControlModifier | Qt.AltModifier | Qt.MetaModifier
          mod = int (e.modifiers () & mask)
          if mod == Qt.ControlModifier :
                cursor.select (QTextCursor.WordUnderCursor)
                if cursor.selectedText() != "" :
                   fmt = cursor.charFormat ()
                   fmt.setFontUnderline (True)
                   fmt.setUnderlineColor (Qt.red)
                   cursor.setCharFormat (fmt)
       super (TextEdit, self).mousePressEvent (e)

   def keyPressEvent (self, e) :
       mask = Qt.ShiftModifier | Qt.ControlModifier | Qt.AltModifier | Qt.MetaModifier
       mod = int (e.modifiers () & mask) # necessary for Ctrl Shift
       key = e.key ()
       if mod == Qt.ControlModifier :
          if key == Qt.Key_F :
             self.findText ()
             return
          if key == Qt.Key_G :
             self.goToLine ()
             return
       if mod == Qt.ShiftModifier :
          if key == Qt.Key_F3 :
             self.findPrev ()
             return
       if mod == 0 :
          if key == Qt.Key_F3 :
             self.findNext ()
             return
       super (TextEdit, self).keyPressEvent (e)

   def contextMenuEvent (self, e) :
       menu = self.createStandardContextMenu ()
       menu.addSeparator ()
       menu.addAction ("My Menu Item")
       menu.exec_ (e.globalPos())

   def jump (self, text) :
       # print ("JUMP", "(" + text + ")")
       fileName = None
       line = None
       column = None

       pattern = '\s*File "([^"]*)", line ([0-9]*)'
       m = re.match (pattern, text)
       if m :
          fileName = m.group (1)
          line = m.group (2)
       else :
          pattern = "([^:]*)(:([0-9]*)(:([0-9]*))?)"
          m = re.match (pattern, text)
          if m :
             fileName = m.group (1)
             line = m.group (3)
             column = m.group (5)

       if fileName != None and fileName != "" :
          if line == None or line == "" :
             line = 0
          else :
             line = int (line)
          if column == None or column == "" :
             column = 0
          else :
             column = int (column)

          # print ("file:", fileName)
          # print ("line:",  line)
          # print ("column:", column)

          if self.win != None :
             self.win.jumpToSource (fileName, line, column)

# --------------------------------------------------------------------------

class Editor (QWidget) :
   def __init__ (self, parent = None) :
       super (QWidget, self).__init__ (parent)

       self.edit = TextEdit ()
       self.edit.setReadOnly (True)

       self.layout = QVBoxLayout ()
       self.setLayout (self.layout)

       self.layout.addWidget (self.edit)

       self.edit.findBox.hide ()
       self.layout.addWidget (self.edit.findBox)

       self.edit.lineBox.hide ()
       self.layout.addWidget (self.edit.lineBox)

# --------------------------------------------------------------------------

class Description (object) :
   def __init__ (self) :
      self.defn_ref = None

      self.invoke_doc = QTextDocument ()
      self.entry_doc = QTextDocument ()
      self.stencil_doc = QTextDocument ()
      self.result_doc = QTextDocument ()

      self.invoke_cursor = QTextCursor (self.invoke_doc)
      self.entry_cursor = QTextCursor (self.entry_doc)
      self.stencil_cursor = QTextCursor (self.stencil_doc)
      self.result_cursor = QTextCursor (self.result_doc)

      self.in_parameters = False

# --------------------------------------------------------------------------

class MainWin (QMainWindow) :

   def __init__ (self, parent = None) :
       super (MainWin, self).__init__ (parent)

       self.red = QColor ("red")
       self.blue = QColor ("blue")
       self.green = QColor ("green")
       self.orange = QColor ("orange")
       self.brown = QColor ("peru")
       self.gray = QColor ("gray")
       self.cornflowerblue = QColor ("cornflowerblue")
       self.lime = QColor ("lime")

       self.input_dict = { } # input files --> edit windows

       self.description_stack = [ ]

       self.current = None # current tree item

       self.save_out = None # original standard output
       self.save_err = None

       self.lexer = None

       # self.setStatusBar (QStatusBar ())
       self.resize (640, 480)

       # widgets

       self.tree = QTreeWidget ()
       self.tree.setHeaderHidden (True)
       self.tree.setSelectionMode (QAbstractItemView.SingleSelection)
       # self.tree.setSelectionBehavior (QAbstractItemView.SelectRows)
       self.tree.itemActivated.connect (self.onItemActivated)
       # self.tree.itemClicked.connect (self.onItemActivated)

       self.symbols = QTreeWidget ()
       self.symbols.setHeaderHidden (True)

       self.inputTabs = QTabWidget ()
       # self.inputTabs.tabBar().setFocusPolicy (Qt.NoFocus)

       # self.input = Editor ()
       # self.inputTabs.addTab (self.input, "source")

       self.output = Editor ()
       self.output.edit.win = self

       self.messages = Editor ()
       self.messages.edit.win = self
       self.messages.edit.enableJump = True

       self.define = Editor ()
       self.define.edit.win = self

       self.invoke = Editor ()
       self.invoke.edit.win = self

       self.entry = Editor ()
       self.entry.edit.win = self

       self.stencil = Editor ()
       self.stencil.edit.win = self

       self.result = Editor ()
       self.result.edit.win = self

       # layout

       self.setCentralWidget (self.inputTabs)

       self.setCorner (Qt.TopLeftCorner, Qt.LeftDockWidgetArea)
       self.setCorner (Qt.BottomLeftCorner, Qt.LeftDockWidgetArea)
       self.setCorner (Qt.TopRightCorner, Qt.RightDockWidgetArea)
       self.setCorner (Qt.BottomRightCorner, Qt.RightDockWidgetArea)

       self.symbolDock = QDockWidget ("symbols", self)
       self.symbolDock.setWidget (self.symbols)
       self.addDockWidget (Qt.LeftDockWidgetArea, self.symbolDock)

       self.treeDock = QDockWidget ("tree", self)
       self.treeDock.setWidget (self.tree)
       # self.addDockWidget (Qt.LeftDockWidgetArea, self.treeDock)
       self.tabifyDockWidget (self.symbolDock, self.treeDock)

       self.outputDock = QDockWidget ("output", self)
       self.outputDock.setWidget (self.output)
       self.addDockWidget (Qt.BottomDockWidgetArea, self.outputDock)

       self.messagesDock = QDockWidget ("messages", self)
       self.messagesDock.setWidget (self.messages)
       # self.tabifyDockWidget (self.outputDock, self.messagesDock)
       self.addDockWidget (Qt.BottomDockWidgetArea, self.messagesDock)

       self.defineDock = QDockWidget ("definition", self)
       self.defineDock.setWidget (self.define)
       self.addDockWidget (Qt.RightDockWidgetArea, self.defineDock)

       self.invokeDock = QDockWidget ("invocation", self)
       self.invokeDock.setWidget (self.invoke)
       self.addDockWidget (Qt.RightDockWidgetArea, self.invokeDock)

       self.entryDock = QDockWidget ("entry", self)
       self.entryDock.setWidget (self.entry)
       self.addDockWidget (Qt.RightDockWidgetArea, self.entryDock)

       self.stencilDock = QDockWidget ("stencil", self)
       self.stencilDock.setWidget (self.stencil)
       self.addDockWidget (Qt.RightDockWidgetArea, self.stencilDock)

       self.resultDock = QDockWidget ("result", self)
       self.resultDock.setWidget (self.result)
       self.addDockWidget (Qt.RightDockWidgetArea, self.resultDock)

       # file menu

       self.fileMenu = self.menuBar().addMenu ("&File")

       act = QAction ("&Quit", self)
       act.setShortcut ("Ctrl+Q")
       act.triggered.connect (self.close)
       self.fileMenu.addAction (act)

       # edit menu

       self.editMenu = self.menuBar().addMenu ("&Edit")

       # view menu

       self.viewMenu = self.menuBar().addMenu ("&View")

       for dock in [self.symbolDock,
                    self.treeDock,
                    self.outputDock,
                    self.messagesDock,
                    self.defineDock,
                    self.invokeDock,
                    self.entryDock,
                    self.stencilDock,
                    self.resultDock ] :
           if dock == self.defineDock or dock == self.outputDock :
              self.viewMenu.addSeparator ()
           self.viewMenu.addAction (dock.toggleViewAction ())

   # -- redirect output --

   def write (self, text) :
       if self.save_out != None :
          self.save_out.write (text)
       self.messages.edit.insertPlainText (text)

   def redirectOutput (self) :
       self.save_out = sys.stdout
       self.save_err = sys.stderr

       sys.stdout = self
       sys.stderr = self

   def restoreOutput (self) :
       sys.stdout = self.save_out
       sys.stderr = self.save_err

   # -- tree --

   def openBranch (self, text) :
       item = QTreeWidgetItem ()
       item.setText (0, text)
       if self.current == None :
          self.tree.addTopLevelItem (item)
       else :
          self.current.addChild (item)
       self.current = item
       return item

   def closeBranch (self) :
       if self.current != None :
          self.current = self.current.parent ()

   def addItem (self, text) :
       item = self.openBranch (text)
       self.closeBranch ()
       return item

   def setColor (self, item, color) :
       item.setForeground (0, color)

   def expandAll (self, node) :
       node.setExpanded (True)
       cnt = node.childCount ()
       inx = 0
       while inx < cnt :
          t = node.child (inx)
          self.expandAll (t)
          inx = inx + 1

   def onItemActivated (self, node, column) :
       mask = Qt.ShiftModifier | Qt.ControlModifier | Qt.AltModifier | Qt.MetaModifier
       mod = int (QApplication.keyboardModifiers () & mask)
       # print ("mode", "{:x}".format (mod))
       #if mod == Qt.ControlModifier :
          #print ("CONTROL CLICK")
          #if hasattr (node, "fileInx") and hasattr (node, "lineNum") :
             #editor = self.input_dict.get (node.fileInx)
             #if editor != None :
                #self.inputTabs.setCurrentWidget (editor)
                #editor.edit.selectLine (node.lineNum)
          #if hasattr (node, "outputLine") :
             #self.output.edit.selectLine (node.outputLine)
       if mod == Qt.ControlModifier :
          print ("CONTROL CLICK")
          t = node
          cnt = t.childCount ()
          while cnt > 0 :
             t.setExpanded (True)
             t = t.child (0) # expand first
             cnt = t.childCount ()
       elif mod == Qt.ShiftModifier :
          print ("SHIFT CLICK")
          t = node
          cnt = t.childCount ()
          while cnt > 0 :
             t.setExpanded (True) # expand last
             t = t.child (cnt-1)
             cnt = t.childCount ()
       elif mod == int (Qt.ShiftModifier | Qt.ControlModifier ):
          print ("CONTROL SHIFT CLICK")
          self.expandAll (node)
       else :
          print ("CLICK")
          if hasattr (node, "about") :
             self.click (node.about)
          if hasattr (node, "description") :
             self.showExpansion (node.description)

   # -- clikc --

   def click (self, about) :
       # print ("CLICK", about, type (about))

       self.tree.selectionModel ().clearSelection ()
       cnt = self.tree.topLevelItemCount ()
       for inx in range (cnt) :
           branch = self.tree.topLevelItem (inx)
           self.clickBranch (branch, about)

       for inx in self.input_dict :
          editor = self.input_dict [inx]
          answer = self.clickEdit (editor.edit, about)
          if answer :
             self.inputTabs.setCurrentWidget (editor)

       self.clickEdit (self.output.edit, about)

       # self.clickEdit (self.define.edit, about)
       # self.clickEdit (self.invoke.edit, about)
       # self.clickEdit (self.entry.edit, about)
       # self.clickEdit (self.stencil.edit, about)
       # self.clickEdit (self.result.edit, about)

   def clickBranch (self, item, about) :
       # print ("ITEM", item.text (0))
       if hasattr (item, "about") and item.about == about :
          # print ("FOUND", item.about)
          item.setSelected (True)
          self.tree.scrollToItem (item)
          if hasattr (item, "description") :
             self.showExpansion (item.description)
          return
       cnt = item.childCount ()
       for inx in range (cnt) :
           subitem = item.child (inx)
           self.clickBranch (subitem, about)

   def clickEdit (self, edit, about) :
       cursor = edit.textCursor ()
       cursor.movePosition (QTextCursor.Start)
       while True :
          format = cursor.charFormat ()
          anchor = format.anchorHref ()
          # if anchor != "" :
          #    print ("anchor", anchor)
          if anchor == str (about) :
             # print ("ANCHOR", anchor)
             cursor.movePosition (QTextCursor.StartOfLine)
             cursor.movePosition (QTextCursor.EndOfLine, QTextCursor.KeepAnchor)
             edit.setTextCursor (cursor)
             edit.ensureCursorVisible ()
             return True
          if cursor.atEnd () : break
          cursor.movePosition (QTextCursor.NextCharacter)
       return False

   # -- source files --

   def addSource (self, fileInx, fullName, source) :
       editor = Editor ()
       editor.edit.win = self
       editor.edit.setPlainText (source)

       name = os.path.basename (fullName)
       inx = self.inputTabs.addTab (editor, name)
       self.inputTabs.setTabToolTip (inx, fullName)

       self.input_dict [fileInx] = editor
       return editor

   def openSource (self) :
       input = self.lexer.input
       fileInx = input.fileInx
       if fileInx not in self.input_dict :
          fullName = input.getFileName ()
          source = input.getSource ()
          self.addSource (fileInx, fullName, source)
       editor = self.input_dict [fileInx]
       self.inputTabs.setCurrentWidget (editor)

   def jumpToSource (self, fileName, line = 0, column = 0) :
       fullName = os.path.abspath (fileName)
       fileInx = fileNameToIndex (fullName)
       if fileInx in self.input_dict :
          editor = self.input_dict [fileInx]
       else :
          f = open (fullName, "r")
          source = f.read ()
          f.close ()
          editor = self.addSource (fileInx, fullName, source)
       self.inputTabs.setCurrentWidget (editor)
       editor.edit.selectLine (line)

   # -- symbols --

   def showSymbols (self) :
       self.symbols.clear ()
       if self.lexer != None :
          for name in sorted (self.lexer.macros) :
              defn = self.lexer.macros [name]
              text = defn.name + " = " + defn.code
              item = QTreeWidgetItem ()
              item.setText (0, text)
              self.symbols.addTopLevelItem (item)

   # -- utilities --

   def showPosition (self) :
       "show token position in source window"
       if self.lexer != None :
          input = self.input_dict.get (self.lexer.tokenFileInx)
          if input != None :
             self.inputTabs.setCurrentWidget (input)
             edit = input.edit
             cursor = edit.textCursor ()
             cursor.setPosition (self.lexer.charByteOfs)
             cursor.movePosition (QTextCursor.StartOfLine)
             cursor.movePosition (QTextCursor.EndOfLine, QTextCursor.KeepAnchor)
             edit.setTextCursor (cursor)
             edit.ensureCursorVisible ()

   def showStatus (self, text) :
       print (text)
       QCoreApplication.processEvents ()

   # def markPosition (self, format, fileInx, lineNum, colNum, byteOfs) :
   #     text = str (fileInx) + "," + str (lineNum) + "," + str (colNum) + "," + str (byteOfs)
   #     format.setAnchorHref (text)
   #     format.setToolTip ("position " + text)

   def markDirective (self, color, about = -1) :
       "set color in source window"
       lexer = self.lexer
       input = self.input_dict.get (lexer.charFileInx)
       if input != None :
          edit = input.edit
          cursor = edit.textCursor ()
          cursor.setPosition (self.directive_location)
          cursor.setPosition (lexer.charByteOfs, QTextCursor.KeepAnchor)
          format = cursor.charFormat ()
          format.setForeground (color)
          if about >= 0 :
             format.setAnchorHref (str (about))
          # self.markPosition (format, lexer.charFileInx, lexer.charLineNum, lexer.charColNum, lexer.charByteOfs)
          cursor.setCharFormat (format)

   def markToken (self, color, about) :
       "set color in source window"
       lexer = self.lexer
       input = self.input_dict.get (lexer.tokenFileInx)
       if input != None :
          edit = input.edit
          cursor = edit.textCursor ()
          cursor.setPosition (lexer.tokenByteOfs)
          cursor.setPosition (lexer.charByteOfs, QTextCursor.KeepAnchor)
          format = cursor.charFormat ()
          format.setForeground (color)
          if about >= 0 :
             format.setAnchorHref (str (about))
             format.setToolTip ("About: " + str (about))
          # self.markPosition (format, lexer.tokenFileInx, lexer.tokenLineNum, lexer.tokenColNum, lexer.tokenByteOfs)
          cursor.setCharFormat (format)

   def markExpansion (self, defn, about) :
       "set color in source window"
       lexer = self.lexer
       input = self.input_dict.get (lexer.tokenFileInx)
       if input != None :
          edit = input.edit
          cursor = edit.textCursor ()
          cursor.setPosition (self.lexer.tokenByteOfs)
          cursor.setPosition (self.lexer.charByteOfs, QTextCursor.KeepAnchor)
          format = cursor.charFormat ()
          format.setForeground (self.red)
          format.setAnchorHref (str (about))
          format.setToolTip ("About: " + str (about) + ", Expand " + defn.name)
          # self.markPosition (format, lexer.tokenFileInx, lexer.tokenLineNum, lexer.tokenColNum, lexer.tokenByteOfs)
          cursor.setCharFormat (format)

   def markComment (self) :
       "set comment color in source window"
       lexer = self.lexer
       input = self.input_dict.get (lexer.charFileInx)
       if input != None :
          edit = input.edit
          cursor = edit.textCursor ()
          cursor.setPosition (self.comment_location)
          cursor.setPosition (lexer.charByteOfs, QTextCursor.KeepAnchor)
          format = cursor.charFormat ()
          format.setForeground (self.gray)
          cursor.setCharFormat (format)

   def putToken (self, cursor, color, about) :
       "add token to output/result window"
       format = cursor.charFormat ()
       format.setAnchorHref (str (about))
       format.setToolTip ("About: " + str (about))
       format.setForeground (color)
       cursor.setCharFormat (format)
       cursor.insertText (lexer.tokenText)

   def treePosition (self, item, char_pos = False) :
       "add file name, line number, tooltip, about and output position to tree"
       if char_pos :
          item.fileInx = lexer.charFileInx
          item.lineNum = lexer.charLineNum
          item.colNum = lexer.charColNum
          item.byteOfs = lexer.charByteOfs
       else :
          item.fileInx = lexer.tokenFileInx
          item.lineNum = lexer.tokenLineNum
          item.colNum = lexer.tokenColNum
          item.byteOfs = lexer.tokenByteOfs

       text = ( lexer.getFileName () +
                ", line: " + str (item.lineNum) +
                ", col: " + str (item.colNum) +
                ", byte: " + str (item.byteOfs) )

       if hasattr (item, "about") :
          text = text + ", about: " + str (item.about)

       item.setToolTip (0, text)

       pos = self.output.edit.textCursor().blockNumber()+1
       item.outputLine = pos

   def treeToken (self, text, color, about) :
       "add token information into tree"
       lexer = self.lexer
       if lexer.token == lexer.eos :
          name = "< end of source >"
       else :
          name = lexer.tokenText

       item = self.addItem (text + " " + name)
       self.setColor (item, color)

       item.about = about
       self.treePosition (item)

   def treeDirective (self, item, color) :
       "add directive into tree (and set color in source window)"
       self.about_count = self.about_count + 1
       about = self.about_count

       # also mark directive in source text
       self.markDirective (color, about)

       self.setColor (item, color)
       item.about = about
       self.treePosition (item, char_pos = True)

   # -- follow --

   def start (self, lexer) :
       self.directive_location = 0
       self.comment_location = 0
       self.about_count = 0
       self.lexer = lexer
       self.openSource ()
       # item = self.addItem ("start")

   def oneCharacter (self) :
       if 0 :
          lexer = self.lexer
          c = lexer.ch
          if c >= ' ' :
             txt = str (c)
          else :
             txt = "ascii (" + str (ord (c)) + ")"
          item = self.addItem ("char " + txt)
          item.setForeground (0, self.cornflowerblue)
          self.treePosition (item, char_pos = True)

   def whiteSpace (self) :
       if 0 :
          c = self.lexer.ch
          edit = self.output.edit
          cursor = edit.textCursor ()
          format = cursor.charFormat ()
          format.setForeground (self.gray)
          cursor.setCharFormat (format)
          if c == '\n' :
             cursor.insertText ("\\n\n")
          elif c == '\r' :
             cursor.insertText ("\\r")
          else :
             cursor.insertText (".")

   def startComment (self, delta) :
       self.comment_location = self.lexer.charByteOfs - delta

   def comment (self) :
       if 0 :
          self.markComment ()

   def unusedToken (self) :
       if 0 :
          self.about_count = self.about_count + 1
          about = self.about_count

          self.markToken (self.gray, about)
          self.treeToken ("unused token", self.gray, about)

   def oneToken (self) :
       if 1 :
          self.about_count = self.about_count + 1
          about = self.about_count

          # tree
          if 0 :
             if self.lexer.isIdentifier () and self.lexer.tokenText.startswith ('@') :
                self.treeToken ("", self.orange, about)
             else :
                self.treeToken ("token", self.brown, about)

          # source
          if 0 :
             self.markToken (self.brown, about)

          # output
          if 0 :
             self.putToken (self.output.edit.textCursor(), self.brown, about)

          # macro expansion
          if len (self.description_stack) != 0 :
             desc = self.description_stack [-1]
             self.putToken (desc.stencil_cursor, self.brown, about)

          for desc in self.description_stack :
             self.putToken (desc.result_cursor, self.brown, about)

   def startDirective (self) :
       self.directive_location = self.lexer.tokenByteOfs

   def endDirective (self) :
       pass

   def unusedDirective (self) :
       if 0 :
          item = self.addItem ("unused directive")
          self.treeDirective (item, self.gray)

   def ifDirective (self) :
       if 1 :
          item = self.openBranch ("#if" + " " + str (self.lexer.use_input))
          self.treeDirective (item, self.green)

   def ifdefDirective (self) :
       if 1 :
          item = self.openBranch ("#ifdef" + " " + str (self.lexer.use_input))
          self.treeDirective (item, self.green)

   def ifndefDirective (self) :
       if 1 :
          item = self.openBranch ("#ifndef" + " " + str (self.lexer.use_input))
          self.treeDirective (item, self.green)

   def elifDirective (self) :
       if 1 :
          self.closeBranch ()
          item = self.openBranch ("#elif" + " " + str (self.lexer.use_input))
          self.treeDirective (item, self.green)

   def elseDirective (self) :
       if 1 :
          self.closeBranch ()
          item = self.openBranch ("#else" + " " + str (self.lexer.use_input))
          self.treeDirective (item, self.green)

   def endifDirective (self) :
       if 1 :
          self.closeBranch ()
          item = self.addItem ("#endif")
          self.treeDirective (item, self.green)

   def evalDirective (self, text) :
       if 1 :
          item = self.addItem ("#eval " + text)
          self.treeDirective (item, self.orange)

   def printDirective (self, text) :
       if 1 :
          item = self.addItem ("#print " + text)
          self.treeDirective (item, self.orange)

   def warningDirective (self, text) :
       if 1 :
          item = self.addItem ("#warning " + text)
          self.treeDirective (item, self.orange)

   def errorDirective (self, text) :
       if 1 :
          item = self.addItem ("#error " + text)
          self.treeDirective (item, self.red)

   def defineDirective (self, defn) :
       if 1 :
          item = self.openBranch ("#define " + defn.name)
          self.treeDirective (item, self.blue)
          if defn.params != None :
             for param in defn.params :
                item = self.addItem ("parameter " + param)
                self.setColor (item, self.blue)
          self.closeBranch ()

   def undefDirective (self, name) :
       if 1 :
          item = self.openBranch ("#undef " + name)
          self.treeDirective (item, self.blue)
          self.closeBranch ()

   def openInclude (self) :
       if 1 :
          self.openSource ()
          item = self.openBranch ("#include " + self.lexer.getFileName ())
          self.treeDirective (item, self.green)
          self.showStatus (self.lexer.getFileName ())

   def closeInclude (self) :
       if 1 :
          self.closeBranch ()
          self.showStatus (self.lexer.getFileName ())

   def expandMacro (self, defn) :
       if 0 :
          self.about_count = self.about_count + 1
          about = self.about_count

          item = self.openBranch ("expand " + defn.name)
          self.setColor (item, self.red)
          item.about = about
          self.treePosition (item)

          if len (self.description_stack) != 0 :
             desc = self.description_stack [-1]
             self.putToken (desc.stencil_cursor, self.red, about)

          item.description = Description ()
          item.description.defn_ref = defn
          self.description_stack.append (item.description)
          self.putToken (item.description.invoke_cursor, self.blue, about)
          self.putToken (item.description.entry_cursor, self.blue, about)

          self.markExpansion (defn, about)

   def closeExpand (self) :
       if 0 :
          self.closeBranch ()
          self.description_stack.pop ()

   def openParameters (self) :
       if 0 :
          item = self.openBranch ("parameters")
          self.setColor (item, self.red)

          if len (self.description_stack) != 0 :
             desc = self.description_stack [-1]
             desc.in_parameters = True

   def closeParameters (self) :
       if 0 :
          self.closeBranch ()

          if len (self.description_stack) != 0 :
             desc = self.description_stack [-1]
             desc.in_parameters = False

   def openParam (self, param) :
       if 0 :
          item = self.openBranch ("parameter " + param)
          self.setColor (item, self.red)
          self.treePosition (item)

   def paramToken (self) :
       if 0 :
          self.about_count = self.about_count + 1
          about = self.about_count

          # tree
          if 1 :
             self.treeToken ("parameter token", self.brown, about)

          # source
          if 1 :
             self.markToken (self.brown, about)

          # macro expansion
          if len (self.description_stack) != 0 :
             desc = self.description_stack [-1]
             self.putToken (desc.invoke_cursor, self.brown, about)
             self.putToken (desc.entry_cursor, self.brown, about)

   def closeParam (self) :
       if 0 :
          self.closeBranch ()

   # -- expansion --

   def showExpansion (self, desc) :

       defn = desc.defn_ref
       code = "#define " + defn.name
       if defn.params != None :
          code = code + "("
          any = False
          for param in defn.params :
              if any :
                 code = code + ","
              any = True
              code = code + param
          code = code + ")"
       code = code + " " + defn.code
       self.define.edit.setPlainText (code)

       self.invoke.edit.setDocument (desc.invoke_doc)
       self.entry.edit.setDocument (desc.entry_doc)
       self.stencil.edit.setDocument (desc.stencil_doc)
       self.result.edit.setDocument (desc.result_doc)

   # -- compare --

   def compare (self, lexer) :
       toFirst = False
       toSecond = False
       first = [ ]
       second = [ ]

       while not lexer.isEndOfSource () :
          if lexer.isKeyword ("@clear") :
             lexer.macros = { }
          elif lexer.isKeyword ("@start") :
             toFirst = True
             toSecond = False
             first = [ ]
             item = self.openBranch ("first")
             self.setColor (item, self.orange)
          elif lexer.isKeyword ("@compare") :
             toFirst = False
             toSecond = True
             second = [ ]
             self.closeBranch()
             item = self.openBranch ("second")
             self.setColor (item, self.orange)
          elif lexer.isKeyword ("@stop") :
             toFirst = False
             toSecond = False
             self.closeBranch()
             print ("FIRST", first)
             print ("SECOND", second)
             ok = len (first) == len (second)
             if ok :
                for inx in range (len (first)) :
                   if first [inx] != second [inx] :
                      ok = False
             if ok :
                item = self.addItem ("O.K.")
                self.setColor (item, self.lime)
             else  :
                item = self.addItem ("BAD")
                self.setColor (item, self.red)
                lexer.warning ("BAD")
          else :
             if toFirst :
                first.append (lexer.tokenText)
             if toSecond :
                second.append (lexer.tokenText)

          # if lexer.tokenText.startswith ("#") :
          #    print ("token", lexer.token, lexer.tokenText)
          lexer.nextToken ()

# --------------------------------------------------------------------------

app = QApplication (sys.argv)
QIcon.setThemeName ("oxygen") # /usr/share/icons/oxygen
win = MainWin ()
win.show ()

win.redirectOutput ()
win.resize (800, 600)
try:
   lexer = Lexer ()
   lexer.follow = win
   lexer.at_keywords = True

   example_dir = "examples/"
   lexer.addInclDir (example_dir)

   if 0 :
      lexer.openFile (example_dir + "example-ifdef.cc")

   if 0 :
      lexer.openFile (example_dir + "example.cc")

   if 0 :
      lexer.openFile (example_dir + "example-preproc.cc")

   if 0 :
      lexer.tolerate_missing_include = True
      lexer.readOptions (gcc_options ())
      lexer.readOptions (pkg_options ("gtk+-2.0"))
      lexer.openFile (example_dir + "example2.cc")

   if 0 :
      lexer.tolerate_missing_include = True
      lexer.readOptions (gcc_options ())
      lexer.addMacro ("__linux__", "")
      lexer.addMacro ("__GNUC__", "4")
      lexer.readOptions (pkg_options ("qt-mt"))
      lexer.openFile (example_dir + "example3.cc")

   if 0 :
      # lexer.tolerate_missing_include = True
      # lexer.addMacro ("__linux__", "")
      # lexer.addMacro ("__GNUC__", "4")
      # lexer.addInclDir ("/usr/include")
      lexer.readOptions (gcc_options ())
      lexer.readOptions (pkg_options ("QtGui"))
      lexer.openFile (example_dir + "example4.cc")

   if 0 :
      lexer.tolerate_missing_include = True
      # lexer.readOptions (gcc_options ())
      lexer.readOptions (pkg_options ("Qt5Widgets"))
      lexer.openFile (example_dir + "example5.cc")

   if 0 :
      lexer.readOptions (gcc_options ())
      lexer.openFile (example_dir + "example6.cc")

   win.compare (lexer)

   # while not lexer.isEndOfSource () :
   #    lexer.nextToken ()

# except LexerException as e:
except Exception as e:
   # print ("Exception", e.args)
   win.showPosition ()
   traceback.print_exc ()
   if isinstance (e, LexerException) :
      print (e.args[0])
win.restoreOutput ()
win.showSymbols ()

app.exec_ ()

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
