
# cecko3_compiler.py

from __future__ import print_function

from cecko3_parser import *

# --------------------------------------------------------------------------

class Compiler (Parser) :

   def __init__ (self) :
       super (Compiler, self).__init__ ()
       self.display = [ ]

   # initialization

   def on_start_program (self, decl_list) :
       decl_list.item_name = ""
       decl_list.qual_name = ""
       decl_list.item_dict = { }
       decl_list.item_list = [ ]
       self.display.append (decl_list)

   # scope

   def enter (self, item) :
       top = self.display [-1]
       top.item_dict [item.item_name] = item
       top.item_list.append (item)
       item.item_context = top
       if top.qual_name == "" :
          item.qual_name = item.item_name
       else :
          item.qual_name = top.qual_name + "." + item.item_name

   def openScope (self, item) :
       if not hasattr (item, "item_dict") :
          item.item_dict = { }
       if not hasattr (item, "item_list") :
          item.item_list = [ ]
       self.display.append (item)

   def closeScope (self) :
       self.display.pop ()

   def lookup (self, name) :
       result = None
       inx = len (self.display) - 1
       while result == None and inx >= 0 :
          result = self.display [inx].item_dict.get (name)
          inx = inx - 1
       return result

   # namespace

   def open_namespace (self, ns) :
       ns.item_name = ns.name
       ns.item_expand = True
       ns.icon = "namespace"
       self.enter (ns)
       self.markDefn (ns)

       self.openScope (ns)
       self.openCompletion (ns, prev = False)

   def close_namespace (self, ns) :
       self.closeCompletion (ns, prev = True)
       self.closeScope ()

   # class

   def begin_class (self, cls) :
       self.openRegion (cls) # region with background color

   def open_class (self, cls) :
       cls.item_name = cls.name
       cls.item_expand = True
       cls.icon = "class"
       self.enter (cls) # add declaration into current scope
       self.markDefn (cls) # mark as definition (in editor)

       self.openScope (cls) # open scope for identifiers
       self.openCompletion (cls, prev = False) # area with text completion

   def close_class (self, cls) :
       self.closeCompletion (cls, prev = True)
       self.closeScope ()

   def end_class (self, cls) :
       self.closeRegion ()

   # enumeration type

   def open_enum (self, typ) :
       typ.item_name = typ.name
       typ.item_expand = True
       typ.icon = "enum"
       self.enter (typ) # add declaration into current scope
       self.markDefn (typ) # mark as definition (in editor)

       self.openScope (typ) # open scope for identifiers
       self.openCompletion (typ, prev = False) # area with text completion

   def close_enum (self, typ) :
       self.closeCompletion (typ, prev = True)
       self.closeScope ()

   def on_enum_item (self, decl) :
       decl.item_name = decl.name
       decl.icon = "variable"
       self.enter (decl)
       self.markDefn (decl)

   # simple declaration

   def on_simple_decl (self, decl) :
       decl.item_name = decl.name
       decl.icon = "variable"
       self.enter (decl)
       self.markDefn (decl)

   def on_local_decl (self, stat) :
       self.on_simple_decl (stat.inner_decl)

   def open_parameters (self, decl) :
       self.openScope (decl)

   def on_param_decl (self, decl) :
       self.on_simple_decl (decl)

   def close_parameters (self, decl) :
       self.closeScope ()

   def open_function (self, decl) :
       decl.icon = "function"
       self.openScope (decl)
       self.openRegion (decl)
       self.openCompletion (decl, prev = False)

   def close_function (self, decl) :
       self.closeCompletion (decl, prev = True)
       self.closeRegion ()
       self.closeScope ()

   # expression

   def on_variable_expr (self, expr) :
       expr.item_name = expr.name
       # print ("ident_expr", expr.item_name)
       expr.item_decl = self.lookup (expr.item_name)
       if expr.item_decl != None :
          expr.qual_name = expr.item_decl.qual_name
       self.markUsage (expr)

# --------------------------------------------------------------------------

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
